@@ -18,5 +18,5 @@ t_quad = shgpy.make_tensor_real(t_quad)
 save_filename = 'fform/None-None-D_6h(001)-particularized.p'
 
 start = time.time()
-shgpy.fformgen.generate_contracted_fourier_transforms(save_filename, 'uft/ufttheta', t_dipole, t_quad)
+shgpy.fformgen.generate_contracted_fourier_transforms(save_filename, 'uft/ufttheta', t_dipole, t_quad, ndigits=9)
 mylogger.debug(f'Finished CFT generation. Took {time.time()-start} seconds.')
