@@ -28,10 +28,32 @@ import os
 import shutil
 from pathlib import Path
 import ctypes
+from math import ceil
 
 _logger = logging.getLogger(__name__)
 
 
+def _split(a, num_per):
+    ans = []
+    for n in range(ceil(len(a)/num_per)):
+        temp = []
+        for i in range(num_per):
+            try:
+                temp.append(a[n*num_per+i])
+            except IndexError:
+                break
+        ans.append(temp)
+    return ans
+
+
+def _split_expr(expr, num_per):
+    if expr.func != sp.Add:
+        _logger.debug(f'expr has only one term:\n{expr}')
+        return [expr]
+    _logger.debug(f'expr has {len(expr.args)} terms.')
+    return [sp.Add(*g, evaluate=False) for g in _split(expr.args, num_per)]
+
+
 def _check_fform(fform):
     if not fform.get_free_symbols():
         message = (
@@ -92,7 +114,7 @@ def _make_energy_expr_list(fform, fdat, free_symbols=None):
     _logger.debug('Cost expression evaluation took'
                    f' {time.time()-start} seconds.')
 
-    return energy_expr_list
+    return energy_expr_list, xs
 
 
 def _make_denergy_expr(energy_expr):
@@ -101,7 +123,183 @@ def _make_denergy_expr(energy_expr):
     return np.array([sp.diff(energy_expr, xs[i]) for i in range(n)])
 
 
-def _fixed_autowrap(energy_expr, prefix, save_filename=None):
+def _make_model_expr(fform, pc, m, component, free_symbols=None):
+    if free_symbols is None:
+        free_symbols = fform.get_free_symbols()
+
+    M = fform.get_M()
+    xs = sp.MatrixSymbol('xs', len(free_symbols), 1)
+    mapping = {fs:xs[i] for i, fs in enumerate(free_symbols)}
+    mapping[sp.I] = 1j
+    start = time.time()
+    _logger.debug(f'Computing cost function term pc={pc} m={m}')
+    if component == 'real':
+        model_expr = sp.re(fform.get_pc(pc)[n2i(m, M)]).xreplace(mapping)
+    elif component == 'imag':
+        model_expr = sp.im(fform.get_pc(pc)[n2i(m, M)]).xreplace(mapping)
+    _logger.debug('Cost expression evaluation took'
+                   f' {time.time()-start} seconds.')
+    return model_expr
+
+
+def _fixed_autowrap_model(fform, save_folder, free_symbols=None, method='gcc', max_terms_per_file=10):
+
+    codegen = CCodeGen()
+
+    if save_folder is not None:
+        if Path(save_folder).exists():
+            shutil.rmtree(save_folder)
+        Path.mkdir(Path(save_folder))
+
+    cost_func_dict = {}
+    M = fform.get_M()
+
+    for k in fform.get_keys():
+        cost_func_dict[k] = {}
+        for m in np.arange(-M, M+1):
+            cost_func_dict[k][m] = {}
+            for component in ['real', 'imag']:
+                _logger.debug('Writing code for pc={k}, m={m}')
+                full_expr = _make_model_expr(
+                    fform,
+                    k,
+                    m,
+                    component,
+                    free_symbols=free_symbols,
+                )
+                expr_chunks = _split_expr(full_expr, max_terms_per_file)
+                cost_func_dict[k][m][component] = []
+                for ei, expr in enumerate(expr_chunks):
+                    routines = []
+                    prefix = '_'.join((k, str(m), component, str(ei)))
+                    routines.append(
+                        codegen.routine(
+                            'autofunc',
+                            expr,
+                        ),
+                    )
+                    [(c_name, bad_c_code), (h_name, h_code)] = codegen.write(
+                        routines,
+                        prefix,
+                    )
+                    c_code = '#include <complex.h>\n'
+                    c_code += bad_c_code
+                    c_code = c_code.replace('conjugate', 'conj')
+
+                    write_directory = Path(tempfile.mkdtemp()).absolute()
+
+                    c_path = write_directory / Path(prefix + '.c')
+                    h_path = write_directory / Path(prefix + '.h')
+                    o_path = write_directory / Path(prefix + '.o')
+                    so_path = write_directory / Path(prefix + '.so')
+
+                    with open(c_path, 'w') as fh:
+                        fh.write(c_code)
+                        fh.write('\n')
+                    with open(h_path, 'w') as fh:
+                        fh.write(h_code)
+                        fh.write('\n')
+
+                    start = time.time()
+                    _logger.debug(f'Start compiling code for PC={k}, m={m}')
+                    os.system(f'{method} -w -c -lm {c_path} -o {o_path}')
+                    os.system(f'{method} -w -shared {o_path} -o {so_path}')
+                    _logger.debug(f'Compiling code took {time.time()-start} seconds.')
+
+                    if save_folder is not None:
+                        shutil.copy(so_path, Path(save_folder))
+
+                    cost_func_dict[k][m][component].append(_load_func(so_path))
+                    shutil.rmtree(write_directory)
+
+    return cost_func_dict
+
+
+def gen_model_func(fform, save_folder, method='gcc', max_terms_per_file=10):
+    """Generate a model function (a folder of .so files) and save them to disk.
+
+    Parameters
+    ----------
+    fform : fFormContainer
+        The Fourier formula to use
+    save_folder : path_like
+        Folder to save the result.
+    method : str, optional
+        Compiler to use. Tested with gcc and clang. Default is gcc.
+    max_terms_per_file : int, optional
+        If there are too many terms per file, sometimes the compiler fails.
+        This constrains the number of terms per file. Default is 10.
+
+    Returns
+    -------
+    model_func : function
+        Result after compiling generated C code. Signature is
+        ``model_func(xs, pc, m)``, where `xs` is a ``numpy.ndarray`` of floats
+        in the order of ``fform.get_free_symbols``, `pc` is the polarization
+        combination, and `m` is the Fourier component.
+
+    """
+    cost_func_dict = _fixed_autowrap_model(fform, save_folder, method=method, max_terms_per_file=max_terms_per_file)
+
+    def model_func(xs, pc, m):
+        return sum(
+            [
+                v(xs.astype(float)) for v in cost_func_dict[pc][m]['real']
+            ]
+        ) + sum(
+            [
+                1j*v(xs.astype(float)) for v in cost_func_dict[pc][m]['imag']
+            ]
+        )
+
+    return model_func
+
+
+def load_model_func(fform, save_folder):
+    """Load a model function generated by ``gen_model_func``
+
+    Returns
+    -------
+    model_func : function
+        Result after compiling generated C code. Signature is
+        ``model_func(xs, pc, m)``, where `xs` is a ``numpy.ndarray`` of floats
+        in the order of ``fform.get_free_symbols``, `pc` is the polarization
+        combination, and `m` is the Fourier component.
+
+    """
+    cost_func_dict = _load_func_dict(fform, save_folder)
+
+    def model_func(xs, pc, m):
+        return sum(
+            [
+                v(xs.astype(float)) for v in cost_func_dict[pc][m]['real']
+            ]
+        ) + sum(
+            [
+                1j*v(xs.astype(float)) for v in cost_func_dict[pc][m]['imag']
+            ]
+        )
+
+    return model_func
+
+
+def _load_func_dict(fform, save_folder):
+    save_folder = Path(save_folder)
+    cost_func_dict = {}
+    M = fform.get_M()
+    for k in fform.get_keys():
+        cost_func_dict[k] = {}
+        for m in np.arange(-M, M+1):
+            cost_func_dict[k][m] = {}
+            for component in ['real', 'imag']:
+                cost_func_dict[k][m][component] = []
+                for path in save_folder.glob('_'.join((k, str(m), component))+'*.so'):
+                    _logger.debug(f'Loading cost function from {path}')
+                    cost_func_dict[k][m][component].append(_load_func(path))
+    return cost_func_dict
+
+
+def _fixed_autowrap(energy_expr, prefix, save_filename=None, method='gcc'):
 
     codegen = CCodeGen()
 
@@ -134,8 +332,8 @@ def _fixed_autowrap(energy_expr, prefix, save_filename=None):
 
     start = time.time()
     _logger.debug('Start compiling code.')
-    os.system(f'gcc -c -lm {c_path} -o {o_path}')
-    os.system(f'gcc -shared {o_path} -o {so_path}')
+    os.system(f'{method} -w -c -lm {c_path} -o {o_path}')
+    os.system(f'{method} -w -shared {o_path} -o {so_path}')
     _logger.debug(f'Compiling code took {time.time()-start} seconds.')
 
     if save_filename is not None:
@@ -150,7 +348,13 @@ def _fixed_autowrap(energy_expr, prefix, save_filename=None):
     return cost_func
 
 
-def _make_energy_func_chunked(energy_expr_list, prefix, save_filename=None):
+def _make_energy_func_chunked(
+    energy_expr_list,
+    prefix,
+    variable,
+    save_filename=None,
+    method='gcc',
+):
 
     codegen = CCodeGen()
 
@@ -158,7 +362,7 @@ def _make_energy_func_chunked(energy_expr_list, prefix, save_filename=None):
     for i, expr in enumerate(energy_expr_list):
         _logger.debug(f'Writing code for expr {i} of {len(energy_expr_list)}')
         routines.append(
-            codegen.routine('expr'+str(i), expr)
+            codegen.routine('expr'+str(i), expr, argument_sequence=[variable])
         )
     [(c_name, bad_c_code), (h_name, h_code)] = codegen.write(
         routines,
@@ -198,8 +402,8 @@ double autofunc(double *xs){
 
     start = time.time()
     _logger.debug('Start compiling code.')
-    os.system(f'gcc -c -lm -fPIC {c_path} -o {o_path}')
-    os.system(f'gcc -shared -fPIC {o_path} -o {so_path}')
+    os.system(f'{method} -w -c -lm -fPIC {c_path} -o {o_path}')
+    os.system(f'{method} -w -shared -fPIC {o_path} -o {so_path}')
     _logger.debug(f'Compiling code took {time.time()-start} seconds.')
 
     if save_filename is not None:
@@ -214,16 +418,22 @@ double autofunc(double *xs){
     return cost_func
 
 
-def _make_energy_func_auto(energy_expr, save_filename=None):
+def _make_energy_func_auto(energy_expr, save_filename=None, method='gcc'):
     energy_func = _fixed_autowrap(
         energy_expr,
         'SHGPY_COST_FUNC',
         save_filename,
+        method=method,
     )
 
     return energy_func
 
 
+def load_func(load_cost_func_filename):
+    """Load a cost function generated by ``gen_cost_func``"""
+    return _load_func(load_cost_func_filename)
+
+
 def _load_func(load_cost_func_filename):
     start = time.time()
     c_lib = ctypes.CDLL(load_cost_func_filename)
@@ -239,10 +449,11 @@ def _load_func(load_cost_func_filename):
 
 
 def _make_energy_func_wrapper(fform, fdat, free_symbols=None,
-                              chunk=False, save_filename=None):
+                              chunk=False, save_filename=None,
+                              method='gcc'):
 
     if chunk:
-        energy_expr_list = _make_energy_expr_list(
+        energy_expr_list, xs = _make_energy_expr_list(
             fform,
             fdat,
             free_symbols,
@@ -250,7 +461,9 @@ def _make_energy_func_wrapper(fform, fdat, free_symbols=None,
         energy_func = _make_energy_func_chunked(
             energy_expr_list,
             'SHGPY_COST_FUNC',
-            save_filename,
+            variable=xs,
+            save_filename=save_filename,
+            method=method,
         )
         return energy_func
 
@@ -263,12 +476,13 @@ def _make_energy_func_wrapper(fform, fdat, free_symbols=None,
         energy_func = _make_energy_func_auto(
             energy_expr,
             save_filename,
+            method=method,
         ) 
         return energy_func
 
 
 def gen_cost_func(fform, fdat, argument_list=None,
-                  chunk=False, save_filename=None):
+                  chunk=False, save_filename=None, method='gcc'):
     """Generate a cost function as an .so file and save it to disk.
 
     Parameters
@@ -285,15 +499,20 @@ def gen_cost_func(fform, fdat, argument_list=None,
         Fourier component in fform and fdat). Default is False.
     save_filename : path_like, optional
         Filename to save the result. Default is not to save.
+    method : str, optional
+        Compiler to use. Tested with gcc and clang. Default is gcc.
 
     Returns
     -------
     cost_func : function
-        Result after compiling generated C code.
+        Result after compiling generated C code. Signature is
+        ``cost_func(xs)``, where `xs` is a ``numpy.ndarray`` of floats
+        in the order of ``fform.get_free_symbols``.
 
     """
     return _make_energy_func_wrapper(fform, fdat, free_symbols=argument_list,
-                                     chunk=chunk, save_filename=save_filename)
+                                     chunk=chunk, save_filename=save_filename,
+                                     method=method)
         
         
 def _make_denergy_func_auto(denergy_expr, save_filename_prefix=None):
